/// <reference types="node" />
import { EventEmitter } from 'events';
import { Socket } from 'net';
import Debug from 'debug';
import { AddNotificationReqCallback, AdsNotificationTarget, AdsRequest, DeleteNotificationReqCallback, GenericReqCallback, ReadDeviceInfoReqCallback, ReadReqCallback, ReadStateReqCallback, ReadWriteReqCallback, ServerConnection, ServerCoreSettings, WriteControlReqCallback, WriteReqCallback } from './types/ads-server';
import { AdsCommandToSend, AmsHeader, AmsTcpHeader, AmsTcpPacket } from './types/ads-types';
/**
 * Base abstract class for ADS server
 */
export declare abstract class ServerCore extends EventEmitter {
    protected debug: Debug.Debugger;
    protected debugD: Debug.Debugger;
    protected debugIO: Debug.Debugger;
    /**
     * Active settings
     */
    settings: ServerCoreSettings;
    /**
     * Active connection information
     */
    connection: ServerConnection;
    /**
     * Active debug level
     *  - 0 = no debugging
     *  - 1 = Extended exception stack trace
     *  - 2 = basic debugging (same as $env:DEBUG='ads-server')
     *  - 3 = detailed debugging (same as $env:DEBUG='ads-server,ads-server:details')
     *  - 4 = full debugging (same as $env:DEBUG='ads-server,ads-server:details,ads-server:raw-data')
     */
    debugLevel: number;
    /**
     * Next invoke ID to use (for notifications)
     */
    protected nextInvokeId: number;
    /**
     * Callback used for ams/tcp commands (like port register)
     */
    protected amsTcpCallback?: ((packet: AmsTcpPacket) => void);
    /**
     * Callback for AMS router state change
     * Used only with ads-server-router
     */
    protected routerStateChangedCallback?: (data: AmsTcpPacket) => void;
    /**
     * Callbacks for ADS command requests
     */
    protected requestCallbacks: {
        [key: string]: GenericReqCallback;
    };
    /**
     * Settings to use are provided as parameter
     */
    constructor(settings: Partial<ServerCoreSettings>);
    /**
     * Sets debugging using debug package on/off.
     * Another way for environment variable DEBUG:
     *  - 0 = no debugging
     *  - 1 = Extended exception stack trace
     *  - 2 = basic debugging (same as $env:DEBUG='ads-server')
     *  - 3 = detailed debugging (same as $env:DEBUG='ads-server,ads-server:details')
     *  - 4 = full debugging (same as $env:DEBUG='ads-server,ads-server:details,ads-server:raw-data')
     *
     * @param level 0 = none, 1 = extended stack traces, 2 = basic, 3 = detailed, 4 = detailed + raw data
     */
    setDebugging(level: number): void;
    /**
     * Checks received data buffer for full AMS packets. If full packet is found, it is parsed and handled.
     * Calls itself recursively if multiple packets available. Added also setImmediate calls to prevent event loop from blocking
     *
     * @param buffer Data buffer to check
     * @param socket Socket connection to use for responding
     * @param setNewBufferCallback Callback that is called to update current data buffer contents
     */
    protected handleReceivedData(buffer: Buffer, socket: Socket, setNewBufferCallback: (newBuffer: Buffer) => void): void;
    /**
     * Parses an AMS/TCP packet from given buffer and then handles it
     *
     * @param data Buffer that contains data for a single full AMS/TCP packet
     * @param socket Socket connection to use for responding
     */
    protected parseAmsTcpPacket(data: Buffer, socket: Socket): Promise<void>;
    /**
     * Parses an AMS/TCP header from given buffer
     *
     * @param data Buffer that contains data for a single full AMS/TCP packet
     * @returns Object `{amsTcp, data}`, where amsTcp is the parsed header and data is rest of the data
     */
    protected parseAmsTcpHeader(data: Buffer): {
        amsTcp: AmsTcpHeader;
        data: Buffer;
    };
    /**
     * Parses an AMS header from given buffer
     *
     * @param data Buffer that contains data for a single AMS packet (without AMS/TCP header)
     * @returns Object `{ams, data}`, where ams is the parsed AMS header and data is rest of the data
     */
    protected parseAmsHeader(data: Buffer): {
        ams: AmsHeader;
        data: Buffer;
    };
    /**
     * Parses ADS data from given buffer. Uses `packet.ams` to determine the ADS command.
     *
     * @param data Buffer that contains data for a single ADS packet (without AMS/TCP header and AMS header)
     * @returns Object that contains the parsed ADS data
     */
    protected parseAdsData(packet: AmsTcpPacket, data: Buffer): AdsRequest;
    /**
     * Handles the parsed AMS/TCP packet and actions/callbacks etc. related to it.
     *
     * @param packet Fully parsed AMS/TCP packet, includes AMS/TCP header and if available, also AMS header and ADS data
     */
    protected onAmsTcpPacketReceived(packet: AmsTcpPacket, socket: Socket): void;
    /**
     * Handles received ADS command
     *
     * @param packet Fully parsed AMS/TCP packet, includes AMS/TCP header, AMS header and ADS data
     * @param socket Socket connection to use for responding
     */
    protected onAdsCommandReceived(packet: AmsTcpPacket, socket: Socket): void;
    /**
     *
     * @param data ADS command to send
     * @param socket Socket connection to use for responding
     */
    protected sendAdsCommand(data: AdsCommandToSend, socket: Socket): Promise<void>;
    /**
     * Creates an AMS/TCP request from given packet
     *
     * @param packet Object containing the full AMS/TCP packet
     * @returns Full created AMS/TCP request as buffer
     */
    protected createAmsTcpRequest(packet: AmsTcpPacket): Buffer;
    /**
     * Creates an AMS header from given packet
     *
     * @param packet Object containing the full AMS/TCP packet
     * @returns Created AMS header as buffer
     */
    protected createAmsHeader(packet: AmsTcpPacket): Buffer;
    /**
     * Creates an AMS/TCP header from given packet and AMS header
     *
     * @param packet Object containing the full AMS/TCP packet
     * @param amsHeader Buffer containing the previously created AMS header
     * @returns Created AMS/TCP header as buffer
     */
    protected createAmsTcpHeader(packet: AmsTcpPacket, amsHeader: Buffer): Buffer;
    /**
     * Sends a device notification to target.
     * @param notification Contains notification target info
     * @param sourceAdsPort ADS port where notification is sent from
     * @param socket Socket where to send data to
     * @param data Data what to send
     */
    protected sendDeviceNotificationToSocket(notification: AdsNotificationTarget, sourceAdsPort: number, socket: Socket, data: Buffer): Promise<void>;
    /**
     * Sets request callback for given ADS command
     *
     * @param request ADS command as number
     * @param callback Callback function
     */
    protected setRequestCallback(request: number, callback: GenericReqCallback): void;
    /**
     * Writes given message to console if `settings.hideConsoleWarnings` is false
     *
     * @param str String to write to console
     */
    protected consoleWrite(str: string): void;
    /**
     * Sets callback function to be called when ADS Read request is received
     *
     * @param callback Callback that is called when request received
     * ```js
     *  onReadReq(async (req, res) => {
     *    //do something with req object and then respond
     *    await res({..})
     *  })
     * ```
     */
    onReadReq(callback: ReadReqCallback): void;
    /**
     * Sets callback function to be called when ADS ReadWrite request is received
     *
     * @param callback Callback that is called when request received
     * ```js
     *  onReadWriteReq(async (req, res) => {
     *    //do something with req object and then respond
     *    await res({..})
     *  })
     * ```
     */
    onReadWriteReq(callback: ReadWriteReqCallback): void;
    /**
     * Sets callback function to be called when ADS Write request is received
     *
     * @param callback Callback that is called when request received
     * ```js
     *  onWriteReq(async (req, res) => {
     *    //do something with req object and then respond
     *    await res({..})
     *  })
     * ```
     */
    onWriteReq(callback: WriteReqCallback): void;
    /**
     * Sets callback function to be called when ADS ReadDeviceInfo request is received
     *
     * @param callback Callback that is called when request received
     * ```js
     *  onReadDeviceInfo(async (req, res) => {
     *    //do something with req object and then respond
     *    await res({..})
     *  })
     * ```
     */
    onReadDeviceInfo(callback: ReadDeviceInfoReqCallback): void;
    /**
     * Sets callback function to be called when ADS ReadState request is received
     *
     * @param callback Callback that is called when request received
     * ```js
     *  onReadState(async (req, res) => {
     *    //do something with req object and then respond
     *    await res({..})
     *  })
     * ```
     */
    onReadState(callback: ReadStateReqCallback): void;
    /**
     * Sets callback function to be called when ADS AddNotification request is received
     *
     * @param callback Callback that is called when request received
     * ```js
     *  onAddNotification(async (req, res) => {
     *    //do something with req object and then respond
     *    await res({..})
     *  })
     * ```
     */
    onAddNotification(callback: AddNotificationReqCallback<AdsNotificationTarget>): void;
    /**
     * Sets callback function to be called when ADS DeleteNotification request is received
     *
     * @param callback Callback that is called when request received
     * ```js
     *  onDeleteNotification(async (req, res) => {
     *    //do something with req object and then respond
     *    await res({..})
     *  })
     * ```
     */
    onDeleteNotification(callback: DeleteNotificationReqCallback): void;
    /**
     * Sets callback function to be called when ADS WriteControl request is received
     *
     * @param callback Callback that is called when request received
     * ```js
     *  onWriteControl(async (req, res) => {
     *    //do something with req object and then respond
     *    await res({..})
     *  })
     * ```
     */
    onWriteControl(callback: WriteControlReqCallback): void;
    /**
     * Writes data to the socket
     *
     * @param data Data to write
     * @param socket Socket to write to
     */
    protected socketWrite(data: Buffer, socket: Socket): Promise<void>;
    /**
     * Converts array of bytes (or Buffer) to AmsNetId string
     *
     * @param byteArray
     * @returns AmsNetId as string (like 192.168.1.10.1.1)
     */
    protected byteArrayToAmsNedIdStr(byteArray: Buffer | number[]): string;
    /**
     * Converts AmsNetId string to array of bytes
     *
     * @param str AmsNetId as string
     * @returns AmsNetId as array of bytes
     */
    protected amsNetIdStrToByteArray(str: string): number[];
}
/**
 * Own exception class used for Server errors
 *
 * Derived from Error but added innerException and ADS error information
 *
 */
export declare class ServerException extends Error {
    sender: string;
    adsError: boolean;
    adsErrorInfo: Record<string, unknown> | null;
    metaData: unknown | null;
    errorTrace: Array<string>;
    getInnerException: () => (Error | ServerException | null);
    stack: string | undefined;
    constructor(server: ServerCore, sender: string, messageOrError: string | Error | ServerException, ...errData: unknown[]);
}
