/// <reference types="node" />
import { ServerCore } from './ads-server-core';
import { AmsRouterState, RouterServerSettings, ServerConnection, AdsNotificationTarget } from './types/ads-server';
import { AmsTcpPacket } from './types/ads-types';
/**
 * TwinCAT ADS server
 *
 * This ADS server class connects to the AMS/ADS router
 * and listens for incoming ADS commands **(ONLY)** for the ADS port provided in settings.
 *
 * **Requires TwinCAT installation / AMS router for operation.** Without router, see `StandAloneServer` class.
 */
export declare class RouterServer extends ServerCore {
    /**
     * Active settings
     */
    settings: RouterServerSettings;
    /**
     * Local router state (if known)
     */
    routerState?: AmsRouterState;
    /**
     * Received data buffer
     */
    private receiveBuffer;
    /**
     * Socket instance
     */
    private socket?;
    /**
     * Handler for socket error event
     */
    private socketErrorHandler?;
    /**
     * Handler for socket close event
     */
    private socketConnectionLostHandler?;
    /**
     * Timer ID and handle of reconnection timer
     */
    private reconnectionTimer;
    /**
     * Timer handle for port register timeout
     */
    private portRegisterTimeoutTimer?;
    /**
     * Creates a new ADS server instance.
     *
     * Settings are provided as parameter
     */
    constructor(settings: Partial<RouterServerSettings>);
    /**
     * Connects to the AMS router and registers ADS port.
     * Starts listening for incoming ADS commands
     *
     * @returns {ServerConnection} Connection info
     */
    connect(): Promise<ServerConnection>;
    /**
     * Disconnects from target router and unregisters ADS port.
     * Stops listening for incoming ADS commands
     */
    disconnect(): Promise<void>;
    /**
     * Reconnects to the AMS router.
     * First disconnects and then connects again.
     *
     * @returns {ServerConnection} Connection info
     */
    reconnect(): Promise<ServerConnection>;
    /**
     * Sends a given data as notification using given
     * notificationHandle and target info.
     */
    sendDeviceNotification(notification: AdsNotificationTarget, data: Buffer): Promise<void>;
    /**
     * Connects to the AMS router and registers ADS port.
     *
     * @param isReconnecting Not used at the moment (true = reconnecting in progress)
     */
    private connectToTarget;
    /**
     * Unregisters ADS port from router (if it was registered)
     * and disconnects target system and ADS router
     *
     * @param [forceDisconnect] - If true, the connection is dropped immediately (default = false)
     * @param [isReconnecting] - If true, call is made during reconnecting
     */
    private disconnectFromTarget;
    /**
     * Disconnects and reconnects again
     *
     * @param [forceDisconnect] - If true, the connection is dropped immediately (default = false)
     * @param [isReconnecting] - If true, call is made during reconnecting
     */
    private reconnectToTarget;
    /**
     * Registers a new ADS port from AMS router
     */
    private registerAdsPort;
    /**
     * Unregisters previously registered ADS port from AMS router.
     * Connection is usually also closed by remote during unregisterin.
     */
    private unregisterAdsPort;
    /**
     * Event listener for socket errors
     */
    private onSocketError;
    /**
     * Called when connection to the remote is lost
     *
     * @param socketFailure - If true, connection was lost due socket/tcp problem -> Just destroy the socket
     *
     */
    private onConnectionLost;
    /**
     * Called when local AMS router status has changed (Router notification received)
     * For example router state changes when local TwinCAT changes from Config to Run state and vice-versa
     *
     * @param data Packet that contains the new router state
     */
    protected onRouterStateChanged(data: AmsTcpPacket): void;
    /**
     * Clears given timer if it's available and increases the id
     * @param timerObject Timer object {id, timer}
     */
    private clearTimer;
}
