/// <reference types="node" />
import { ServerCore } from './ads-server-core';
import { AddNotificationReqCallback, ServerConnection, StandAloneAdsNotificationTarget, StandAloneServerSettings } from './types/ads-server';
/**
 * TwinCAT ADS server
 *
 * This ADS server starts up a server at provided TCP port (default: ADS.ADS_DEFAULT_TCP_PORT = 48898).
 *
 * Then it listens for incoming ADS commands for **any ADS port** on this AmsNetId.
 *
 * **Does not require TwinCAT installation/AMS router (for example Raspberry Pi, Linux, etc...).**
 * With router, see `Server` class.
 */
export declare class StandAloneServer extends ServerCore {
    /**
     * Active settings
     */
    settings: StandAloneServerSettings;
    /**
     * Next free available connection ID
     */
    private nextConnectionid;
    /**
     * Object containing all active connections
     */
    private connections;
    /**
     * Socket server instance
     */
    private server?;
    /**
     * Settings to use are provided as parameter
     */
    constructor(settings: StandAloneServerSettings);
    /**
     * Setups a listening server and then starts listening for incoming connections
     * and ADS commands.
     */
    listen(): Promise<ServerConnection>;
    /**
     * Stops listening for incoming connections and closes the server.
     * If closing fails, throws an error but the server instance is destroyed anyways.
     * So the connection is always closed after calling close()
     */
    close(): Promise<void>;
    /**
     * Sends a given data as notification using given
     * notificationHandle and target info.
     */
    sendDeviceNotification(notification: StandAloneAdsNotificationTarget, data: Buffer): Promise<void>;
    /**
     * Sets callback function to be called when ADS AddNotification request is received
     *
     * @param callback Callback that is called when request received
     * ```js
     *  onAddNotification(async (req, res) => {
     *    //do something with req object and then respond
     *    await res({..})
     *  })
     * ```
     */
    onAddNotification(callback: AddNotificationReqCallback<StandAloneAdsNotificationTarget>): void;
    /**
     * Called when error is thrown at server instance
     *
     * @param err Error object from this.server
     */
    private onServerError;
    /**
     * Called on a new connection at server instance
     *
     * @param socket The new socket connection that was created
     */
    private onServerConnection;
}
