"use strict";
/*
https://github.com/jisotalo/ads-server

Copyright (c) 2021 Jussi Isotalo <j.isotalo91@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.StandAloneServer = void 0;
const net_1 = require("net");
const ads_server_core_1 = require("./ads-server-core");
const ADS = __importStar(require("./ads-commons"));
/**
 * TwinCAT ADS server
 *
 * This ADS server starts up a server at provided TCP port (default: ADS.ADS_DEFAULT_TCP_PORT = 48898).
 *
 * Then it listens for incoming ADS commands for **any ADS port** on this AmsNetId.
 *
 * **Does not require TwinCAT installation/AMS router (for example Raspberry Pi, Linux, etc...).**
 * With router, see `Server` class.
 */
class StandAloneServer extends ads_server_core_1.ServerCore {
    /**
     * Settings to use are provided as parameter
     */
    constructor(settings) {
        super(settings);
        /**
         * Active settings
         */
        this.settings = {
            listeningTcpPort: ADS.ADS_DEFAULT_TCP_PORT,
            localAmsNetId: ''
        };
        /**
         * Next free available connection ID
         */
        this.nextConnectionid = 0;
        /**
         * Object containing all active connections
         */
        this.connections = {};
        /**
         * Socket server instance
         */
        this.server = undefined;
        //Taking the default settings and then updating the provided ones
        this.settings = {
            ...this.settings,
            ...settings
        };
    }
    /**
     * Setups a listening server and then starts listening for incoming connections
     * and ADS commands.
     */
    listen() {
        return new Promise(async (resolve, reject) => {
            try {
                this.debug(`listen(): Creating socket server`);
                this.server = net_1.createServer(this.onServerConnection.bind(this));
                //Error during startup
                const handleErrorAtStart = (err) => {
                    this.debug(`listen(): Creating socket server failed: ${err.message}`);
                    reject(err);
                };
                this.server.on('error', handleErrorAtStart);
                //Starting to listening
                this.server.listen(this.settings.listeningTcpPort, this.settings.listeningAddress, () => {
                    var _a, _b, _c;
                    this.debug(`listen(): Listening on ${((_a = this.server) === null || _a === void 0 ? void 0 : _a.address()).port}`);
                    (_b = this.server) === null || _b === void 0 ? void 0 : _b.off('error', handleErrorAtStart);
                    (_c = this.server) === null || _c === void 0 ? void 0 : _c.on('error', this.onServerError.bind(this));
                    this.connection = {
                        connected: true,
                        localAmsNetId: this.settings.localAmsNetId
                    };
                    resolve(this.connection);
                });
            }
            catch (err) {
                this.connection.connected = false;
                reject(err);
            }
        });
    }
    /**
     * Stops listening for incoming connections and closes the server.
     * If closing fails, throws an error but the server instance is destroyed anyways.
     * So the connection is always closed after calling close()
     */
    close() {
        return new Promise(async (resolve, reject) => {
            this.debug(`close(): Closing socket server`);
            this.connection.connected = false;
            if (!this.server) {
                return resolve();
            }
            try {
                this.server.close(err => {
                    var _a;
                    if (err) {
                        this.debug(`close(): Error during closing server but server is destroyed: ${err.message}`);
                        this.server = undefined;
                        return reject(err);
                    }
                    (_a = this.server) === null || _a === void 0 ? void 0 : _a.removeAllListeners();
                    this.server = undefined;
                    resolve(err);
                });
            }
            catch (err) {
                reject(err);
                this.server = undefined;
            }
        });
    }
    /**
     * Sends a given data as notification using given
     * notificationHandle and target info.
     */
    sendDeviceNotification(notification, data) {
        if (!this.connection.connected)
            throw new ads_server_core_1.ServerException(this, 'sendDeviceNotification()', `Server is not active. Use listen() first.`);
        if (!notification.socket)
            throw new ads_server_core_1.ServerException(this, 'sendDeviceNotification()', `Required notification.socket is missing.`);
        return this.sendDeviceNotificationToSocket(notification, notification.sourceAdsPort, notification.socket, data);
    }
    /**
     * Sets callback function to be called when ADS AddNotification request is received
     *
     * @param callback Callback that is called when request received
     * ```js
     *  onAddNotification(async (req, res) => {
     *    //do something with req object and then respond
     *    await res({..})
     *  })
     * ```
     */
    onAddNotification(callback) {
        this.setRequestCallback(ADS.ADS_COMMAND.AddNotification, callback);
    }
    /**
     * Called when error is thrown at server instance
     *
     * @param err Error object from this.server
     */
    onServerError(err) {
        this.debug(`onServerError(): Socket server error, disconnecting: ${err.message}`);
        this.emit("server-error", err);
        //Closing connection
        this.close()
            .catch();
    }
    /**
     * Called on a new connection at server instance
     *
     * @param socket The new socket connection that was created
     */
    onServerConnection(socket) {
        this.debug(`onServerConnection(): New connection from ${socket.remoteAddress}`);
        //Creating own data object for this connection
        const id = this.nextConnectionid++;
        const connection = {
            id,
            socket,
            buffer: Buffer.alloc(0)
        };
        //Handling incoming data
        socket.on('data', (data) => {
            if (this.debugIO.enabled) {
                this.debugIO(`IO in  <------ ${data.byteLength} bytes from ${socket.remoteAddress}: ${data.toString('hex')}`);
            }
            else if (this.debugD.enabled) {
                this.debugD(`IO in  <------ ${data.byteLength} bytes from ${socket.remoteAddress}`);
            }
            //Adding received data to connection buffer and checking the data
            connection.buffer = Buffer.concat([connection.buffer, data]);
            this.handleReceivedData(connection.buffer, connection.socket, (newBuffer) => connection.buffer = newBuffer);
        });
        //Handling socket closing
        socket.on("close", (hadError) => {
            this.debug(`onServerConnection(): Connection from ${socket === null || socket === void 0 ? void 0 : socket.remoteAddress} was closed (error: ${hadError})`);
            //Removing the connection
            socket === null || socket === void 0 ? void 0 : socket.removeAllListeners();
            delete this.connections[id];
        });
        socket.on("end", () => {
            this.debug(`onServerConnection(): Connection from ${socket === null || socket === void 0 ? void 0 : socket.remoteAddress} was ended`);
            //Removing the connection
            socket === null || socket === void 0 ? void 0 : socket.removeAllListeners();
            delete this.connections[id];
        });
        socket.on('error', (err) => {
            this.debug(`onServerConnection(): Connection from ${socket.remoteAddress} had error: ${err}`);
        });
        this.connections[id] = connection;
    }
}
exports.StandAloneServer = StandAloneServer;
